#!/usr/bin/env python
from datetime import datetime, timedelta
from functools import wraps

import logging
import os
import time
import yaml

import numpy as np
import pandas as pd
import xarray as xr

from mapies import MAPIES
from util.func_tools import *
#timeit, get_file_list, time_converter, frequency_int, error_estimation, time_domain_selection, geo_to_rot
from grids.monarch import RotatedGrid, RegularGrid
from pathlib import Path
from glob import glob as glob_module

import multiprocessing

os.environ["OMP_NUM_THREADS"] = "8"


# TODO: still need to do a load of filtering huuuhhh
# TODO: Create Yaml config file for time, lat, lon, uncertainty const etc.


# Uncertainty constants should be able to be passed as number, array or dict

def rotate_function(r, lon, lat, obs):

    if isinstance(r, RotatedGrid):
        # Aggregate the observations to the grid representation
        lon_agg, lat_agg, rlon_agg, rlat_agg, obs_agg = r.aggregate(lon, lat, obs)
        
        return lon_agg, lat_agg, obs_agg
    
    elif isinstance(r, RegularGrid):
        lon_agg, lat_agg, obs_agg = r.aggregate(lon, lat, obs)

        return lon_agg, lat_agg, obs_agg
    
    else:
        raise ValueError("Invalid grid representation")

def process_batch(r, batch_files, obs_var, lat_var, lon_var, chunk_size):
    """
    Process a batch of files to calculate cumulative sums and counts.

    Parameters:
    - batch_files: List of file paths in the batch.
    - obs_var, lat_var, lon_var: Variable names for observations, latitude, and longitude.
    - rotate_func: Function to apply rotation and aggregation.

    Returns:
    - cumulative_obs, cumulative_lat, cumulative_lon, cumulative_count: Aggregated results for the batch.
    """
    def preprocess(ds):
        """Add a unique dimension to distinguish datasets."""
        return ds.expand_dims(dim={"index": [ds.attrs["product_name"]]})

    # Open batch of files as a multi-file dataset
    # t8 = time.perf_counter()
    ds = xr.open_mfdataset(
        batch_files,
        chunks={"index": chunk_size},
        parallel=True,
        engine="h5netcdf",
        preprocess=preprocess,
    )
    # t9 = time.perf_counter()
    # print(f'Opened batch in time {t9-t8}', flush=True)

    # Flatten data arrays
    obs = ds[obs_var].values.flatten()
    lat = ds[lat_var].values.flatten()
    lon = ds[lon_var].values.flatten()

    # Rotate and aggregate
    t1  = time.perf_counter()
    lon_agg, lat_agg, obs_agg = rotate_function(r,lon, lat, obs)
    t2 = time.perf_counter()
    print(f'Finished rotation in time {t2-t1}', flush=True)

    valid_mask = ~np.isnan(obs_agg)
    cumulative_obs = np.zeros_like(obs_agg)
    cumulative_lat = np.zeros_like(lat_agg)
    cumulative_lon = np.zeros_like(lon_agg)
    cumulative_count = np.zeros_like(obs_agg)

    # Update cumulative arrays
    cumulative_obs[valid_mask] = obs_agg[valid_mask]
    cumulative_lat[valid_mask] = lat_agg[valid_mask]
    cumulative_lon[valid_mask] = lon_agg[valid_mask]
    cumulative_count[valid_mask] = 1

    ds.close()
    # Debugging: Ensure all outputs are correct
    # print(f"Batch results - obs: {cumulative_obs.shape}, lat: {cumulative_lat.shape}, lon: {cumulative_lon.shape}, count: {cumulative_count.shape}", flush=True)

    return cumulative_obs, cumulative_lat, cumulative_lon, cumulative_count



class VIIRS(MAPIES):
    """
    Class for VIIRS specific data
    """

    def __init__(self, start_date, end_date, **kwargs):
        """
        Inherited init class with new variables
        """
        super().__init__(start_date, end_date)
        
        self.time_orig = np.datetime64("1993-01-01T00:00:00")
        self.datatype="viirs"
        frequency = kwargs.get("frequency")
        self.dest = kwargs.get("dest")
        self.indir = kwargs.get("indir")

        # =============================================================================
        # CARLOTTA ADDED THIS
        # =============================================================================
        # self.year = kwargs.get("year", 2024)  # Default to 2024
        # self.months = kwargs.get("months", list(range(1, 13)))  # Default to all months (1-12)
        # self.monthly_files = {}  # Dictionary to store files by month

        self.year = int(start_date[:4])  
        self.start_month = int(start_date[4:6])  # Extract start month from start_date
        self.end_month = int(end_date[4:6])  # Extract end month from end_date
        self.months = list(range(self.start_month, self.end_month + 1))  # Dictionary to store files by month
        self.start_day = int(start_date[6:8])  
        self.end_day = int(end_date[6:8]) 
        self.monthly_files = {}  # Dictionary to store files by month
        # =============================================================================


        self.dates_slice = pd.date_range(
            self.start_date, 
            self.end_date, 
            freq=frequency
            ).strftime('%Y%m%d%H%M')
        if frequency:
            self.frequency = frequency_int(frequency)

        if isinstance(self.start_date, str):
            self.start_date = time_converter(self.start_date)

        if isinstance(self.end_date, str):
            self.end_date = time_converter(self.end_date)

        self.read_config(**kwargs)


    @timeit
    def read_config(self, **kwargs):
        """
        Read yaml config file
        """
        module_dir = os.path.dirname(__file__)


        # Let the user read in their own config
        config_file = kwargs.get("config_file")
        if config_file:
            try:
                config = yaml.safe_load(open(config_file))
            except FileNotFoundError:
                logging.error("This is not a config file")
        else:
            config = yaml.safe_load(open(os.path.join(module_dir, "config/satellite_config.yaml")))


        variable_dict = config[self.datatype]["variables"]
        self.time_var = variable_dict["time_variable"]
        self.lon_var = variable_dict["lon_variable"]
        self.lat_var = variable_dict["lat_variable"]

        # If obs_var passed then run data analysis on that variable if not pull default
        obs_var = kwargs.get("obs_var")
        if obs_var:
            self.obs_var = obs_var
        else:
            self.obs_var = variable_dict["obs_variable"]


        da_dict = config[self.datatype]["da"]
        self.obsid_da = da_dict["obsid"] #int

        unc_const = kwargs.get("unc_const")
        if unc_const:
            self.unc_const = unc_const
        else:
            self.unc_const = da_dict["uncertainty_constants"] #float, int, list, dict
        

        grid_repr = kwargs.get("grid_repr")
        if grid_repr:
            self.grid_repr = grid_repr
        else:
            self.grid_repr = da_dict["grid_repr"] #str



    @timeit
    def preprocess_vars(self):
        """
        Preprocessing of the dataset
        """
        super().preprocess_vars()

        # Lon and lat values
        lon_dims = self.ds[self.lon_var].dims
        lon_shape = self.ds[self.lon_var].shape
        lon_attrs = self.ds[self.lon_var].attrs
        self.lon_values = self.ds[self.lon_var].values.flatten()

        # Reindex lon column values
        self.lon_values = self.reindex(
            self.time_values_index, 
            self.lon_values,
            )

        lat_dims = self.ds[self.lat_var].dims
        lat_shape = self.ds[self.lat_var].shape
        lat_attrs = self.ds[self.lat_var].attrs
        self.lat_values = self.ds[self.lat_var].values.flatten()

        # Reindex aod column values
        self.lat_values = self.reindex(
            self.time_values_index, 
            self.lat_values,
            )


        # AOD column values, default "Aerosol_Optical_Thickness_550_Land_Ocean_Best_Estimate"
        aod_dims = self.ds[self.obs_var].dims
        aod_shape = self.ds[self.obs_var].shape
        aod_attrs = self.ds[self.obs_var].attrs
        self.obs = self.ds[self.obs_var].values.flatten()

        # Reindex aod column values
        self.obs = self.reindex(
            self.time_values_index, 
            self.obs,
            )


    @timeit
    def rotate(self):
        """
        Perform Rotation of Grid representation
        """
        # Calculate rotated lon and lat values 
        
        # Calculate the grid representation
        r = RotatedGrid(centre_lon=20, centre_lat=35, dlon=.1, dlat=.1, west=-51, south=-35)
        # Aggregate the the observations to the grid representation
        lon, lat, rlon, rlat, obs = r.aggregate(self.lon_values, self.lat_values, self.obs)
        self.lon_values = lon
        self.lat_values = lat
        self.obs = obs

    @timeit
    def read_nc(self):
        """
        Read netcdf files with xarray
        """
        # Take in start and end date and convert to julian dates
        file_patterns = []
        for date in self.dates_slice:
            date = datetime.strptime(date, '%Y%m%d%H%M').strftime('%Y%j')
            filepaths = f'{self.indir}/{date[0:4]}/{date[4:]}/AERDB_L2_VIIRS_NOAA20.A{date}*'
            print(filepaths)
            file_patterns.append(filepaths)

        # TODO:  maybe change this to NETCDF4
        files = get_file_list(file_patterns)

        # A bit of preprocessing to start before reading in the files
        def preprocess(ds):
            ds = ds.expand_dims(dim={"index": [ds.attrs["product_name"]]})
            return ds

        # Open dataset with xarray and dask
        self.ds = xr.open_mfdataset(files, preprocess=preprocess)

    @timeit
    def to_da(self):
        """
        Function that returns all the needed variables for the DA
        """

        # TODO: Move this
        r = RotatedGrid(centre_lon=20, centre_lat=35, dlon=.1, dlat=.1, west=-51, south=-35)
        # Calculate error
        self.obserr = error_estimation(self.datatype, self.obs, self.unc_const)

        outfiles = []

        # Cut to the frequencies that we want
        # Look at start and end date - Find date ranges in between with that frequency 
        # Then find the boundaries
        # We take the start date as the midpoint of the lower boundary
        for date in self.dates_slice:
            l_border = (datetime.strptime(date, "%Y%m%d%H%M") - timedelta(hours=self.frequency/2)).strftime('%Y%m%d%H%M')
            r_border = (datetime.strptime(date, "%Y%m%d%H%M") + timedelta(hours=self.frequency/2)).strftime('%Y%m%d%H%M')
            # Included by Guillaume in his DA function so I believe he needs this to be true
            filename = Path(self.dest).joinpath(f'obs{date}.nc')
            # Cut down the time interval again then reindex obs, obserr, lat and lon
            # Returning also arrays for type, lev, id and n with the same index as the time var
            time_values, frequency_index = time_domain_selection(
                self.time_values, 
                l_border, 
                r_border,
                )


            # Reindex
            obs = self.reindex(frequency_index, self.obs)
            lon = self.reindex(frequency_index, self.lon_values)
            lat = self.reindex(frequency_index, self.lat_values)
            obserr = self.reindex(frequency_index, self.obserr)
            # Run aggregation with grid representation
            lon, lat, obs, obserr = r.aggregate(lon, lat, obs, obserr)
            rlon, rlat = geo_to_rot(lon, lat, centre_lon=20, centre_lat=35)


            # Create new arrays with same length as obs
            obsid = np.full(shape=obs.shape, fill_value=self.obsid_da, dtype=self.int_type)
            obstype = np.full(shape=obs.shape, fill_value=30, dtype=self.int_type)
            obslev = np.full(shape=obs.shape, fill_value=-99999, dtype=self.int_type)
            obsn = np.full(shape=obs.shape, fill_value=1, dtype=self.int_type)
            time_values = np.full(shape=obs.shape, fill_value=date, dtype=str)

            # Coords equals index

            assert obs.shape == obserr.shape
            assert lon.shape == time_values.shape
            coords = dict(index=("index", np.indices(obs.shape)[0, :]))
            data_vars = dict(
                time=(["index",], time_values),
                lon=(["index",], lon), 
                lat=(["index",], lat),
                rlon=(["index",], rlon), 
                rlat=(["index",], rlat), #Also return rlon/rlat values if available
                obs=(["index",], obs),
                obserr=(["index", ], obserr),
                obsid=(["index", ], obsid),
                obstype=(["index", ], obstype),
                obslev=(["index", ], obslev),
                obsn=(["index", ], obsn),
            )

            ds = self.to_xarray(coords=coords, data_vars=data_vars)

            ds.to_netcdf(filename, encoding={})
            outfiles.append(filename)

        return outfiles

    # Plots
    def plot_2D_obs(self, outdir="./", **kwargs):
        """
        Plotting the observations specific to VIIRS
        """
        if self.grid_repr == "rotated":
            self.rotate()
        super().plot_2D_obs()

    @timeit
    def plot_2D_average(self, month, outdir, batch_size = 10, chunk_size = 1):
        """
        Plot the 2D average for a given month.
        """

        if not hasattr(self, 'monthly_files') or month not in self.monthly_files:
            print(f"No files found for month {month}. Please run discover_files first.")
            return
        print(f'Processing average data for month {month}')

        lon, lat, obs = self.process_monthly_data2(month, batch_size=batch_size, chunk_size=chunk_size)
        # lon, lat, obs  = self.process_average_data3(month, batch_size=50)

        # title = f"Average AOD Plot of Day {self.start_day} of Month {month:02d}"
        title = f"Average AOD Plot of Month {month:02d}"
        filename = f"{outdir}/{self.datatype}_2D_avg_{month:02d}_05res_global_1day_b50_c10.png"
        super().plot_2D_obs_custom(
            lon_values=lon,
            lat_values=lat,
            obs_values=obs,
            outdir=outdir,
            title=title,
            filename=filename,
        )


    def julian_to_date(self, julian_date):
        """
        Convert Julian date to month.

        Parameters:
        - julian_date: Julian date string (e.g., "001" for January 1st).

        Returns:
        - int: Month as an integer (1-12).
        """
        date = datetime.strptime(f"{self.year}{julian_date}", "%Y%j")
        return date.month, date.day
    
    @timeit
    def discover_files(self):
        """
        Discover files for the specified year and months.
        """
        # Construct the path for the specified year
        base_path = os.path.join(self.indir, str(self.year))

        if not os.path.exists(base_path):
            print(f"Directory not found: {base_path}")
            return

        for julian_date in sorted(os.listdir(base_path)):

            if julian_date.isdigit():        
                month, day = self.julian_to_date(julian_date)

                if (
                    self.start_month <= month <= self.end_month and 
                    self.start_day <= day <= self.end_day):
                    subfolder = os.path.join(base_path, julian_date)
                    files = glob_module(os.path.join(subfolder, "*.nc"))  
                    if files:
                        if month not in self.monthly_files:
                            self.monthly_files[month] = []
                        self.monthly_files[month].extend(files)

        print(f"Discovered files for months: {list(self.monthly_files.keys())}")
        for month, files in self.monthly_files.items():
            print(f"  Month {month:02d}: {len(files)} files")


    @timeit
    def process_monthly_data2(self, month, batch_size, chunk_size):
        """
        Process and compute the 2D monthly average for a given month/list of files using multiprocessing.

        Parameters:
        - month: int, the month to process (e.g., 1 for January).
        - batch_size: int, the number of files to process in each batch.
        - chunk_size: int, the size of chunks for processing within each batch.

        Returns:
        - lon_final, lat_final, obs_final: Aggregated longitude, latitude, and observation arrays.
        """
        # r = RotatedGrid(centre_lon=20, centre_lat=35, dlon=.1, dlat=.1, west=-51, south=-35)
        # r = RotatedGrid(centre_lon=0, centre_lat=0, dlat=0.1, dlon=0.1, west=-180, south=-90)
        r = RegularGrid(centre_lon=0, centre_lat=0, dlat=0.5, dlon=0.5, west=-180, south=-90)
        month_files = self.monthly_files[month]
        print(f"Processing {len(month_files)} files for month {month}")

        # Split files into batches
        batches = [
            month_files[i:i + batch_size] for i in range(0, len(month_files), batch_size)
        ]
        args = [(r, batch, self.obs_var, self.lat_var, self.lon_var, chunk_size) for batch in batches]

        # Use multiprocessing to process batches
        with multiprocessing.Pool(processes=8) as pool:
            results = pool.starmap(process_batch, args)

        # print(f"Results from pool.starmap: {results}")

        # Initialize cumulative arrays
        cumulative_obs = None
        cumulative_lat = None
        cumulative_lon = None
        cumulative_count = None

        t3 = time.perf_counter()
        for obs, lat, lon, count in results:
            if cumulative_obs is None:
                # Initialize with the shape of the first result
                cumulative_obs = obs
                cumulative_lat = lat
                cumulative_lon = lon
                cumulative_count = count
            else:
                # Incrementally sum arrays
                cumulative_obs += obs
                cumulative_lat += lat
                cumulative_lon += lon
                cumulative_count += count
        t4 = time.perf_counter()
        print(f"Time to sum arrays: {t4 - t3}")

        t5 = time.perf_counter()
        # Compute final averages
        valid_mask = cumulative_count > 0
        # print(f"Valid mask count: {np.sum(valid_mask)}")
        final_obs = np.zeros_like(cumulative_obs)
        final_lat = np.zeros_like(cumulative_lat)
        final_lon = np.zeros_like(cumulative_lon)

        final_obs[valid_mask] = cumulative_obs[valid_mask] / cumulative_count[valid_mask]
        final_lat[valid_mask] = cumulative_lat[valid_mask] / cumulative_count[valid_mask]
        final_lon[valid_mask] = cumulative_lon[valid_mask] / cumulative_count[valid_mask]
        
        t6 = time.perf_counter()
        print(f"Time to compute final averages: {t6 - t5}")
        print(f"Completed processing for month {month}.")
        return final_lon, final_lat, final_obs
